@@ -1,25 +1,25 @@
 /*
  *   Copyright 2013 Martin Gräßlin <mgraesslin@kde.org>
  *
- *   This program is free software; you can redistribute it and/or modify
- *   it under the terms of the GNU Library General Public License as
- *   published by the Free Software Foundation; either version 2, or
- *   (at your option) any later version.
+ *   This library is free software; you can redistribute it and/or
+ *   modify it under the terms of the GNU Lesser General Public
+ *   License as published by the Free Software Foundation; either
+ *   version 2.1 of the License, or (at your option) any later version.
  *
- *   This program is distributed in the hope that it will be useful,
+ *   This library is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *   GNU General Public License for more details
+ *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *   Lesser General Public License for more details.
  *
- *   You should have received a copy of the GNU Library General Public
- *   License along with this program; if not, write to the
- *   Free Software Foundation, Inc.,
- *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+ *   You should have received a copy of the GNU Lesser General Public
+ *   License along with this library.  If not, see <http://www.gnu.org/licenses/>.
  */
 #include "nettesthelper.h"
 #include <netwm.h>
 #include <qtest_widgets.h>
 #include <QProcess>
+// system
+#include <unistd.h>
 
 Q_DECLARE_METATYPE(NET::State)
 Q_DECLARE_METATYPE(NET::States)
@@ -28,7 +28,7 @@ Q_DECLARE_METATYPE(NET::Actions)
 class Property : public QScopedPointer<xcb_get_property_reply_t, QScopedPointerPodDeleter>
 {
 public:
-    Property(xcb_get_property_reply_t *p = 0) : QScopedPointer<xcb_get_property_reply_t, QScopedPointerPodDeleter>(p) {}
+    Property(xcb_get_property_reply_t *p = nullptr) : QScopedPointer<xcb_get_property_reply_t, QScopedPointerPodDeleter>(p) {}
 };
 
 #define INFO NETWinInfo info(m_connection, m_testWindow, m_rootWindow, NET::WMAllProperties, NET::WM2AllProperties, NET::WindowManager);
@@ -41,7 +41,7 @@ public:
 #define GETPROP(type, length, formatSize) \
     xcb_get_property_cookie_t cookie = xcb_get_property_unchecked(connection(), false, m_testWindow, \
                                        atom, type, 0, length); \
-    Property reply(xcb_get_property_reply(connection(), cookie, Q_NULLPTR)); \
+    Property reply(xcb_get_property_reply(connection(), cookie, nullptr)); \
     QVERIFY(!reply.isNull()); \
     QCOMPARE(reply->format, uint8_t(formatSize)); \
     QCOMPARE(reply->value_len, uint32_t(length));
@@ -49,7 +49,7 @@ public:
 #define VERIFYDELETED(t) \
     xcb_get_property_cookie_t cookieDeleted = xcb_get_property_unchecked(connection(), false, m_testWindow, \
             atom, t, 0, 1); \
-    Property replyDeleted(xcb_get_property_reply(connection(), cookieDeleted, Q_NULLPTR)); \
+    Property replyDeleted(xcb_get_property_reply(connection(), cookieDeleted, nullptr)); \
     QVERIFY(!replyDeleted.isNull()); \
     QVERIFY(replyDeleted->type == XCB_ATOM_NONE);
 
@@ -75,6 +75,7 @@ private Q_SLOTS:
     void testFrameExtents();
     void testFrameExtentsKDE();
     void testFrameOverlap();
+    void testFullscreenMonitors();
 
 private:
     bool hasAtomFlag(const xcb_atom_t *atoms, int atomsLenght, const QByteArray &actionName);
@@ -89,6 +90,7 @@ private:
     QScopedPointer<QProcess> m_xvfb;
     xcb_window_t m_rootWindow;
     xcb_window_t m_testWindow;
+    QByteArray m_displayNumber;
 };
 
 void NetWinInfoTestWM::initTestCase()
@@ -107,24 +109,36 @@ void NetWinInfoTestWM::cleanupTestCase()
 void NetWinInfoTestWM::init()
 {
     // first reset just to be sure
-    m_connection = Q_NULLPTR;
+    m_connection = nullptr;
     m_rootWindow = XCB_WINDOW_NONE;
     m_testWindow = XCB_WINDOW_NONE;
     // start Xvfb
     m_xvfb.reset(new QProcess);
-    // randomize the display id in [1, 98]
-    // 0 is not used because it conflicts with "normal" X server
-    // 99 is not used because it's used by KDE's CI infrastructure
-    const QString id = QStringLiteral(":") + QString::number((qrand() % 98) + 1);
-    m_xvfb->start(QStringLiteral("Xvfb"), QStringList() << id);
+
+    // use pipe to pass fd to Xvfb to get back the display id
+    int pipeFds[2];
+    QVERIFY(pipe(pipeFds) == 0);
+    m_xvfb->start(QStringLiteral("Xvfb"), QStringList{ QStringLiteral("-displayfd"), QString::number(pipeFds[1]) });
     QVERIFY(m_xvfb->waitForStarted());
     QCOMPARE(m_xvfb->state(), QProcess::Running);
-    // give it some time before we open the X Display
-    QTest::qWait(100);
+
+    // reads from pipe, closes write side
+    close(pipeFds[1]);
+
+    QFile readPipe;
+    QVERIFY(readPipe.open(pipeFds[0], QIODevice::ReadOnly, QFileDevice::AutoCloseHandle));
+    QByteArray displayNumber = readPipe.readLine();
+    readPipe.close();
+
+    displayNumber.prepend(QByteArray(":"));
+    displayNumber.remove(displayNumber.size() -1, 1);
+    m_displayNumber = displayNumber;
+
     // create X connection
     int screen = 0;
-    m_connection = xcb_connect(qPrintable(id), &screen);
+    m_connection = xcb_connect(displayNumber.constData(), &screen);
     QVERIFY(m_connection);
+    QVERIFY(!xcb_connection_has_error(m_connection));
     m_rootWindow = KXUtils::rootWindow(m_connection, screen);
 
     // create test window
@@ -253,7 +267,7 @@ void NetWinInfoTestWM::testAllowedActions()
     ATOM(_NET_WM_ALLOWED_ACTIONS)
     INFO
 
-    QCOMPARE(info.allowedActions(), NET::Actions(0));
+    QCOMPARE(info.allowedActions(), NET::Actions());
     QFETCH(NET::Actions, actions);
     info.setAllowedActions(actions);
     QCOMPARE(info.allowedActions(), actions);
@@ -415,6 +429,7 @@ void NetWinInfoTestWM::testState_data()
     const QByteArray maxHoriz         = QByteArrayLiteral("_NET_WM_STATE_MAXIMIZED_HORZ");
     const QByteArray shaded           = QByteArrayLiteral("_NET_WM_STATE_SHADED");
     const QByteArray skipTaskbar      = QByteArrayLiteral("_NET_WM_STATE_SKIP_TASKBAR");
+    const QByteArray skipSwitcher     = QByteArrayLiteral("_KDE_NET_WM_STATE_SKIP_SWITCHER");
     const QByteArray keepAbove        = QByteArrayLiteral("_NET_WM_STATE_ABOVE");
     const QByteArray staysOnTop       = QByteArrayLiteral("_NET_WM_STATE_STAYS_ON_TOP");
     const QByteArray skipPager        = QByteArrayLiteral("_NET_WM_STATE_SKIP_PAGER");
@@ -436,6 +451,7 @@ void NetWinInfoTestWM::testState_data()
     QTest::newRow("fullScreen")       << NET::States(NET::FullScreen)       << (QVector<QByteArray>() << fullScreen);
     QTest::newRow("keepBelow")        << NET::States(NET::KeepBelow)        << (QVector<QByteArray>() << keepBelow);
     QTest::newRow("demandsAttention") << NET::States(NET::DemandsAttention) << (QVector<QByteArray>() << demandsAttention);
+    QTest::newRow("skipSwitcher")     << NET::States(NET::SkipSwitcher)     << (QVector<QByteArray>() << skipSwitcher);
 
     // TODO: it's possible to be keep above and below at the same time?!?
     QTest::newRow("all") << NET::States(NET::Modal |
@@ -448,11 +464,12 @@ void NetWinInfoTestWM::testState_data()
                                    NET::KeepBelow |
                                    NET::Hidden |
                                    NET::FullScreen |
-                                   NET::DemandsAttention)
+                                   NET::DemandsAttention |
+                                   NET::SkipSwitcher)
                          << (QVector<QByteArray>() << modal << sticky << maxVert << maxHoriz
                              << shaded << skipTaskbar << keepAbove
                              << skipPager << hidden << fullScreen
-                             << keepBelow << demandsAttention << staysOnTop);
+                             << keepBelow << demandsAttention << staysOnTop << skipSwitcher);
 }
 
 void NetWinInfoTestWM::testState()
@@ -461,9 +478,9 @@ void NetWinInfoTestWM::testState()
     ATOM(_NET_WM_STATE)
     INFO
 
-    QCOMPARE(info.state(), NET::States(0));
+    QCOMPARE(info.state(), NET::States());
     QFETCH(NET::States, states);
-    info.setState(states, 0);
+    info.setState(states, NET::States());
     QCOMPARE(info.state(), states);
 
     // compare with the X property
@@ -553,6 +570,88 @@ void NetWinInfoTestWM::testVisibleName()
     QCOMPARE(info.visibleName(), "bar");
 }
 
+class MockWinInfo : public NETWinInfo
+{
+public:
+    MockWinInfo(xcb_connection_t *connection, xcb_window_t window, xcb_window_t rootWindow)
+        : NETWinInfo(connection, window, rootWindow, NET::WMAllProperties, NET::WM2AllProperties, NET::WindowManager)
+    {
+    }
+
+protected:
+    void changeFullscreenMonitors(NETFullscreenMonitors topology) override
+    {
+        setFullscreenMonitors(topology);
+    }
+};
+
+void NetWinInfoTestWM::testFullscreenMonitors()
+{
+    // test case for BUG 391960
+    QVERIFY(connection());
+    const uint32_t maskValues[] = {
+                     XCB_EVENT_MASK_SUBSTRUCTURE_REDIRECT |
+                     XCB_EVENT_MASK_KEY_PRESS |
+                     XCB_EVENT_MASK_PROPERTY_CHANGE |
+                     XCB_EVENT_MASK_COLOR_MAP_CHANGE |
+                     XCB_EVENT_MASK_SUBSTRUCTURE_REDIRECT |
+                     XCB_EVENT_MASK_SUBSTRUCTURE_NOTIFY |
+                     XCB_EVENT_MASK_FOCUS_CHANGE | // For NotifyDetailNone
+                     XCB_EVENT_MASK_EXPOSURE
+    };
+    QScopedPointer<xcb_generic_error_t, QScopedPointerPodDeleter> redirectCheck(xcb_request_check(connection(),
+                                                                        xcb_change_window_attributes_checked(connection(),
+                                                                                                                m_rootWindow,
+                                                                                                                XCB_CW_EVENT_MASK,
+                                                                                                                maskValues)));
+    QVERIFY(redirectCheck.isNull());
+
+    KXUtils::Atom atom(connection(), QByteArrayLiteral("_NET_WM_FULLSCREEN_MONITORS"));
+
+    // create client connection
+    auto clientConnection = xcb_connect(m_displayNumber.constData(), nullptr);
+    QVERIFY(clientConnection);
+    QVERIFY(!xcb_connection_has_error(clientConnection));
+
+    NETWinInfo clientInfo(clientConnection, m_testWindow, m_rootWindow, NET::WMAllProperties, NET::WM2AllProperties);
+    NETFullscreenMonitors topology;
+    topology.top = 1;
+    topology.bottom = 2;
+    topology.left = 3;
+    topology.right = 4;
+    clientInfo.setFullscreenMonitors(topology);
+    xcb_flush(clientConnection);
+
+    MockWinInfo info(connection(), m_testWindow, m_rootWindow);
+
+    while (true) {
+        KXUtils::ScopedCPointer<xcb_generic_event_t> event(xcb_wait_for_event(connection()));
+        if (event.isNull()) {
+            break;
+        }
+        if ((event->response_type & ~0x80) != XCB_CLIENT_MESSAGE) {
+            continue;
+        }
+
+        NET::Properties dirtyProtocols;
+        NET::Properties2 dirtyProtocols2;
+        QCOMPARE(info.fullscreenMonitors().isSet(), false);
+        info.event(event.data(), &dirtyProtocols, &dirtyProtocols2);
+        QCOMPARE(info.fullscreenMonitors().isSet(), true);
+        break;
+    }
+    xcb_flush(connection());
+    // now the property should be updated
+    waitForPropertyChange(&info, atom, NET::Property(0), NET::WM2FullscreenMonitors);
+
+    QCOMPARE(info.fullscreenMonitors().top, 1);
+    QCOMPARE(info.fullscreenMonitors().bottom, 2);
+    QCOMPARE(info.fullscreenMonitors().left, 3);
+    QCOMPARE(info.fullscreenMonitors().right, 4);
+
+    xcb_disconnect(clientConnection);
+}
+
 QTEST_GUILESS_MAIN(NetWinInfoTestWM)
 
 #include "netwininfotestwm.moc"
